@@ -1,18 +1,16 @@
-// =================================================================
-//          #     #                 #     #
-//          ##    #   ####   #####  ##    #  ######   #####
-//          # #   #  #    #  #    # # #   #  #          #
-//          #  #  #  #    #  #    # #  #  #  #####      #
-//          #   # #  #    #  #####  #   # #  #          #
-//          #    ##  #    #  #   #  #    ##  #          #
-//          #     #   ####   #    # #     #  ######     #
+// ==========================================================================
+//     _   _ _ ____            ____          _____
+//    | | | (_)  _ \ ___ _ __ / ___|___  _ _|_   _| __ __ _  ___ ___ _ __
+//    | |_| | | |_) / _ \ '__| |   / _ \| '_ \| || '__/ _` |/ __/ _ \ '__|
+//    |  _  | |  __/  __/ |  | |__| (_) | | | | || | | (_| | (_|  __/ |
+//    |_| |_|_|_|   \___|_|   \____\___/|_| |_|_||_|  \__,_|\___\___|_|
 //
-//       ---   The NorNet Testbed for Multi-Homed Systems  ---
-//                       https://www.nntb.no
-// =================================================================
+//       ---  High-Performance Connectivity Tracer (HiPerConTracer)  ---
+//                 https://www.nntb.no/~dreibh/hipercontracer/
+// ==========================================================================
 //
 // High-Performance Connectivity Tracer (HiPerConTracer)
-// Copyright (C) 2015-2022 by Thomas Dreibholz
+// Copyright (C) 2015-2025 by Thomas Dreibholz
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -30,7 +28,9 @@
 // Contact: dreibh@simula.no
 
 #include "resultswriter.h"
+#include "assure.h"
 #include "logger.h"
+#include "tools.h"
 
 #include <unistd.h>
 
@@ -39,25 +39,31 @@
 #include <boost/iostreams/filtering_streambuf.hpp>
 #include <boost/iostreams/filter/bzip2.hpp>
 #include <boost/iostreams/filter/gzip.hpp>
-// #include <boost/iostreams/filter/lzma.hpp>   -- Currently not provided by the Debian/Ubuntu packages!
-// #include <boost/process/environment.hpp>
+#include <boost/iostreams/filter/lzma.hpp>
+#include <boost/process/environment.hpp>
 
 
 // ###### Constructor #######################################################
-ResultsWriter::ResultsWriter(const std::string&            directory,
+ResultsWriter::ResultsWriter(const std::string&            programID,
+                             const unsigned int            measurementID,
+                             const std::string&            directory,
                              const std::string&            uniqueID,
-                             const std::string&            formatName,
+                             const std::string&            prefix,
                              const unsigned int            transactionLength,
+                             const unsigned int            timestampDepth,
                              const uid_t                   uid,
                              const gid_t                   gid,
                              const ResultsWriterCompressor compressor)
-   : Directory(directory),
-     UniqueID(uniqueID),
-     FormatName(formatName),
+   : ProgramID(programID),
+     MeasurementID(measurementID),
+     Directory(directory),
+     Prefix(prefix),
      TransactionLength(transactionLength),
+     TimestampDepth(timestampDepth),
      UID(uid),
      GID(gid),
-     Compressor(compressor)
+     Compressor(compressor),
+     UniqueID(uniqueID)
 {
    Inserts   = 0;
    SeqNumber = 0;
@@ -71,25 +77,26 @@ ResultsWriter::~ResultsWriter()
 }
 
 
+// ###### Specifc output format #############################################
+void ResultsWriter::specifyOutputFormat(const std::string& outputFormatName,
+                                        const unsigned int outputFormatVersion)
+{
+   OutputFormatName    = outputFormatName;
+   OutputFormatVersion = outputFormatVersion;
+}
+
+
 // ###### Prepare directories ###############################################
 bool ResultsWriter::prepare()
 {
    try {
-      const boost::filesystem::path tempDirectory = Directory / "tmp";
-      boost::filesystem::create_directory(Directory);
-      boost::filesystem::create_directory(tempDirectory);
-      const int r1 = chown(Directory.string().c_str(), UID, GID);
-      const int r2 = chown(tempDirectory.string().c_str(), UID, GID);
-      if(r1 || r2) {
-         HPCT_LOG(warning) << "Setting ownership of " << Directory << " and " << tempDirectory
-                           << " to UID " << UID << ", GID " << GID << " failed: " << strerror(errno);
-      }
+      std::filesystem::create_directory(Directory);
    }
    catch(std::exception const& e) {
       HPCT_LOG(error) << "Unable to prepare directories - " << e.what();
-      return(false);
+      return false;
    }
-   return(changeFile());
+   return changeFile();
 }
 
 
@@ -103,11 +110,11 @@ bool ResultsWriter::changeFile(const bool createNewFile)
       try {
          if(Inserts == 0) {
             // empty file -> just remove it!
-            boost::filesystem::remove(TempFileName);
+            std::filesystem::remove(TempFileName);
          }
          else {
             // file has contents -> move it!
-            boost::filesystem::rename(TempFileName, TargetFileName);
+            std::filesystem::rename(TempFileName, TargetFileName);
          }
       }
       catch(std::exception const& e) {
@@ -122,11 +129,9 @@ bool ResultsWriter::changeFile(const bool createNewFile)
       try {
          const char* extension = "";
          switch(Compressor) {
-            /*
             case XZ:
                extension = ".xz";
              break;
-            */
             case BZip2:
                extension = ".bz2";
              break;
@@ -136,16 +141,34 @@ bool ResultsWriter::changeFile(const bool createNewFile)
             default:
              break;
          }
-         const std::string name = UniqueID + str(boost::format("-%09d.results%s") % SeqNumber % extension);
-         TempFileName   = Directory / "tmp" / name;
-         TargetFileName = Directory / name;
+         const std::string name = UniqueID +
+            str(boost::format("-%09d.hpct%s") % SeqNumber % extension);
+         std::filesystem::path targetPath =
+            Directory /
+               makeDirectoryHierarchy<std::chrono::system_clock::time_point>(
+                  std::filesystem::path(),
+                  name, std::chrono::system_clock::now(),
+                  0, TimestampDepth);
+         TargetFileName = targetPath / name;
+         TempFileName   = TargetFileName;
+         TempFileName += ".tmp";
+         try {
+            std::filesystem::create_directories(targetPath);
+         }
+         catch(std::filesystem::filesystem_error& e) {
+            HPCT_LOG(warning) << "Creating directory hierarchy " << targetPath
+                              << " failed: " << e.what();
+            return false;
+         }
          OutputFile.open(TempFileName.c_str(), std::ios_base::out | std::ios_base::binary);
          switch(Compressor) {
-            /*
-            case XZ:
-               OutputStream.push(boost::iostreams::lzma_compressor());
+            case XZ: {
+               const boost::iostreams::lzma_params params(
+                  boost::iostreams::lzma::default_compression,
+                  std::thread::hardware_concurrency());
+               OutputStream.push(boost::iostreams::lzma_compressor(params));
+              }
              break;
-            */
             case BZip2:
                OutputStream.push(boost::iostreams::bzip2_compressor());
              break;
@@ -157,19 +180,14 @@ bool ResultsWriter::changeFile(const bool createNewFile)
          }
          OutputStream.push(OutputFile);
          OutputCreationTime = std::chrono::steady_clock::now();
-         if( (OutputStream.good()) && (chown(TempFileName.c_str(), UID, GID) != 0) ) {
-            HPCT_LOG(warning) << "Setting ownership of " << TempFileName
-                              << " to UID " << UID << ", GID " << GID
-                              << " failed: " << strerror(errno);
-         }
-         return(OutputStream.good());
+         return OutputStream.good();
       }
       catch(std::exception const& e) {
          HPCT_LOG(error) << "ResultsWriter::changeFile() - " << e.what();
-         return(false);
+         return false;
       }
    }
-   return(true);
+   return true;
 }
 
 
@@ -178,46 +196,59 @@ bool ResultsWriter::mayStartNewTransaction()
 {
    const std::chrono::steady_clock::time_point now = std::chrono::steady_clock::now();
    if(std::chrono::duration_cast<std::chrono::seconds>(now - OutputCreationTime).count() > TransactionLength) {
-      return(changeFile());
+      return changeFile();
    }
-   return(true);
+   return true;
 }
 
 
 // ###### Generate INSERT statement #########################################
 void ResultsWriter::insert(const std::string& tuple)
 {
-   OutputStream << tuple << std::endl;
+   if(__builtin_expect(Inserts == 0, 0)) {
+      if(!OutputFormatName.empty()) {
+         // Write header
+         OutputStream << "#? HPCT "
+                     << OutputFormatName    << " "
+                     << OutputFormatVersion << " "
+                     << ProgramID           << "\n";
+      }
+   }
+   OutputStream << tuple << "\n";
    Inserts++;
 }
 
 
 // ###### Prepare results writer ############################################
 ResultsWriter* ResultsWriter::makeResultsWriter(std::set<ResultsWriter*>&       resultsWriterSet,
+                                                const std::string&              programID,
+                                                const unsigned int              measurementID,
                                                 const boost::asio::ip::address& sourceAddress,
-                                                const std::string&              resultsFormat,
+                                                const std::string&              resultsPrefix,
                                                 const std::string&              resultsDirectory,
                                                 const unsigned int              resultsTransactionLength,
+                                                const unsigned int              resultsTimestampDepth,
                                                 const uid_t                     uid,
                                                 const gid_t                     gid,
                                                 const ResultsWriterCompressor   compressor)
 {
    if(!resultsDirectory.empty()) {
       std::string uniqueID =
-         resultsFormat + "-" +
-         str(boost::format("P%d") % getpid()) + "-" +   /* Better: boost::this_process::get_id() */
+         resultsPrefix + "-" +
+         ((measurementID != 0) ?
+            "#" + std::to_string(measurementID) :
+            "P" + std::to_string(boost::this_process::get_id())) + "-" +
          sourceAddress.to_string() + "-" +
          boost::posix_time::to_iso_string(boost::posix_time::microsec_clock::universal_time());
       replace(uniqueID.begin(), uniqueID.end(), ' ', '-');
 
       ResultsWriter* resultsWriter =
-         new ResultsWriter(resultsDirectory, uniqueID, resultsFormat, resultsTransactionLength,
+         new ResultsWriter(programID, measurementID, resultsDirectory, uniqueID,
+                           resultsPrefix, resultsTransactionLength, resultsTimestampDepth,
                            uid, gid, compressor);
-      if(resultsWriter->prepare() == true) {
-         resultsWriterSet.insert(resultsWriter);
-         return(resultsWriter);
-      }
-      delete resultsWriter;
+      assure(resultsWriter != nullptr);
+      resultsWriterSet.insert(resultsWriter);
+      return resultsWriter;
    }
-   return(nullptr);
+   return nullptr;
 }
